/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright (C) 2012 Chris Mrwald (chris at pl2 dot com)
 * All Rights Reserved.
 * Contributed to and published at the Hunspell project
 * http://hunspell.sourceforge.net.
 *
 * Contributor(s):
 * 
 *  
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** **/

/* History
 * 2012-07-24 Chris Mrwald - created
 *
 */

/* Some notes
 *
 * The purpose of this module is hook into any RichEdit control for spell checking,
 * whether the application has been written in C, C++ or uses MFC.
 *
 * The concept is to hook into a message filter of the control to determine
 * the right time when the next potential word has been entered, and perform
 * the spell check for a certain range of characters and words at a time.
 * 
 * The character limit of a RichEdit control can be set either in your application,
 * or if it exceeds a length of 2k you might want to consider modifying the preset
 * of the define for MAX_STRING_LEN found in stringW32.h.
 * 
 * The code provides an option to load two differnt dictionaries, and to verify words
 * in both of them. The second dictionary is only used if a word could be found
 * in the first dictionary.
 * 
 * By default the dictionaries are looked up in working directory of the application.
 * Set the define for DICT_PATH to any local path convinient for your use.
 * 
 * libhunspell.dll is required to run applciations compiled with this module.
 *
 * This code has been written and compiled with MSVS 2008, VC 9.0.
 * It has been tested on Windows 2000, XP, Vista, Windows 7 and Windows 8,
 * as well as under Wine emulation 1.2 and 1.5 on Mac OSX and Linux.
 * 
 * 2012-07-24 Chris Mrwald
 */


#include "../win_api/config.h"
#include "../win_api/hunspelldll.h"
#include "../hunspell/csutil.hxx"
#include "../parsers/textparser.hxx"

#include <windows.h>
#include <richedit.h>
#include <wchar.h>

#include "SpellChecker.h"
#include "stringW32.h"

#define DICT_PATH             "./"
#define DICT_MAX              2
#define DICT_DISABLED         "DISABLED"
// keystrokes to trigger spell check of a word:
#define END_WORD_WCHARS       L" .:,;_-!?=()&%%$\"+*~#0123456789"


Hunspell*   g_pHunSpell[DICT_MAX]     = { 0, 0 };
TextParser* g_pHSTxtParser[DICT_MAX]  = { 0, 0 };
char*       g_dicEnc[DICT_MAX]        = { 0, 0 };
UINT        g_codePage[DICT_MAX]      = { CP_ACP, CP_ACP };
UINT        g_langNum[DICT_MAX]       = { 0, 0 };
bool        g_availabe[DICT_MAX]      = { false, false };

CHARFORMAT2 g_goodFormat;
CHARFORMAT2 g_urlFormat;
CHARFORMAT2 g_badFormat;

char        g_HSDictname[32]          = { 0 };
char        g_HSDictname2[32]         = { 0 };


// convenient debug messages
_inline void debug_log (char* message, ...)
{
  va_list  	    marker;
  char          msg[1024];

  va_start (marker, message);
  vsnprintf_s (msg, _countof(msg), _TRUNCATE, message, marker);
  va_end (marker);

  OutputDebugStringA (msg);
}


// is the character following a word
_inline bool isEndWordWchar(wchar_t tc)
{
  int k = 0;

  if (tc <= L' ')
    return true;  // a control charcter

  for (k = 0; k < _countof(END_WORD_WCHARS); k++)
    if (tc == END_WORD_WCHARS[k])
      break;
  return (k < _countof(END_WORD_WCHARS)) ? true : false;
}

// count hard line breaks within a text area
_inline int ctHardBreaks(wchar_t* bufW, int fromPos, int toPos)
{
  int ct  = 0;
  int bis = toPos;

  if (bis > (int)wcslen(bufW))
    bis = wcslen(bufW);

  for (int i = 0; i < bis; i++)
    if (bufW[i] == L'\r' && bufW[++i] == L'\n')
      ct++;
  return ct;
}

// count characters in an UTF8 encoded string
_inline int ctCharsUTF8(char* utf8, int bis)
{
  int ct = 0;
  for (int i = 0; i < bis ; ) {
    switch (utf8[i] & 0xf0) {
      case 0xc0:
      case 0xd0: // 2-byte UTF-8 codes
        if ((utf8[i + 1] & 0xc0) == 0x80)
          ++i;
      break;
      case 0xe0: // 3-byte UTF-8 codes
        if ((utf8[i + 1] & 0xc0) == 0x80)
          ++i;
        if ((utf8[i + 1] & 0xc0) == 0x80)
          ++i;
        break;
      case 0xf0: // 4-byte or more bytes UTF-8 codes
        i += 3;
        break;
    }
    i++;
    ct++;
  }
  return ct;
}

_inline int check_spelling (HWND hwnd, MSGFILTER* msgf, int lenTextAround)
{

  if (!g_pHunSpell[0] || !g_pHSTxtParser[0] || !hwnd || !msgf)
    return 0;

  wchar_t     bufW[MAX_STRING_LEN];
  CHARRANGE   prevSel;
  int         ctTypoz = 0;
  POINT       scrollPos;

  SendMessage (hwnd, EM_EXGETSEL, 0, (LPARAM)&prevSel);
  if (prevSel.cpMin != prevSel.cpMax)
    return 0;
  SendMessage (hwnd, WM_SETREDRAW, (WPARAM)FALSE, 0);
  SendMessage (hwnd, EM_GETSCROLLPOS, 0, (LPARAM)&scrollPos);

  // load the text
  TEXTRANGE txtRange;
  memset (&txtRange, 0, sizeof(txtRange));
  memset (bufW, 0, sizeof (bufW));
  txtRange.lpstrText  = bufW;
  txtRange.chrg.cpMin = 0;
  txtRange.chrg.cpMax = _countof(bufW);
  SendMessage (hwnd, EM_GETTEXTRANGE, 0, (LPARAM)&txtRange);

  // we inspect 'lenTextAround' chars before the curent selection,
  // and 'lenTextAround' chars after,
  // considering the previous and next word delimiter to get 'full words' always
  wchar_t* start, *end, *p = &bufW[prevSel.cpMin];
  if (lenTextAround <= 0) {
    start = bufW;
    end   = bufW + wcslen(bufW) + 1;
  } else {
    for (start = p - lenTextAround; start > bufW; start--) {
      if (isEndWordWchar (*start))
        break;
    }
    for (end = p + lenTextAround; end < bufW + (_countof(bufW) - 1); end++)
      if (isEndWordWchar (*start))
        break;
  }
  if (start < bufW)
    start = bufW;
  if (end > bufW + (_countof(bufW) - 1))
    end = bufW + (_countof(bufW) - 1);
  if (start >= end)
    return 0;
  CHARRANGE inspectSel;
  inspectSel.cpMin =  (LONG)(start - bufW);
  inspectSel.cpMax =  (LONG)(end - bufW);
  SendMessage (hwnd, EM_EXSETSEL, 0, (LPARAM)&inspectSel);
  // reset that area, for a new spell check
  SendMessage (hwnd, EM_SETCHARFORMAT, SCF_SELECTION, (LPARAM)&g_goodFormat);

  // start parsing and spellchecking
  char* token   = NULL;
  char* token2  = NULL;
  g_pHSTxtParser[0]->put_line (string_from_unicode_cp (start, g_codePage[0]));
  if (g_pHunSpell[1] && g_pHSTxtParser[1])
    g_pHSTxtParser[1]->put_line (string_from_unicode_cp (start, g_codePage[1]));

  while (true) {
    if (!(token = g_pHSTxtParser[0]->next_token()))
      break;

    CHARRANGE wSel;
    int       ctLine;
    int       tPos        = g_pHSTxtParser[0]->get_tokenpos ();
    if (tPos > inspectSel.cpMax)
      break;

    // url checker
    int uPos = tPos;
    g_pHSTxtParser[0]->get_url(tPos, &uPos);
    if (tPos != uPos) {
      // check for url
      if (tPos != uPos) {
        ctLine = ctHardBreaks (bufW, 0, inspectSel.cpMin + tPos);
        wSel.cpMin = inspectSel.cpMin + tPos - ctLine;
        ctLine = ctHardBreaks (bufW, 0, inspectSel.cpMin + uPos);
        wSel.cpMax = wSel.cpMin + uPos - ctLine;
        SendMessage (hwnd, EM_EXSETSEL, 0, (LPARAM)&wSel);
        SendMessage (hwnd, EM_SETCHARFORMAT, SCF_SELECTION, (LPARAM)&g_urlFormat);
        wSel.cpMin = inspectSel.cpMin + uPos - ctLine;
        wSel.cpMax = wSel.cpMin + uPos - ctLine;
        SendMessage (hwnd, EM_EXSETSEL, 0, (LPARAM)&wSel);
        SendMessage (hwnd, EM_SETCHARFORMAT, SCF_SELECTION, (LPARAM)&g_goodFormat);
      }
      free(token);
      continue;
    }
    // --> end of urlchecker

    // check spelling:
    bool bGood1 = hunspell_spell(g_pHunSpell[0], token) ? true : false;

    // check spelling in dict2
    bool bGood2       = false;
    bool bSkipToken2  = true;
    int  tPos2;
    if (g_pHunSpell[1] && g_pHSTxtParser[1]) {
      // try to snycronize token2 to token
      if (!token2) {
        token2  = g_pHSTxtParser[1]->next_token();
        tPos2   = g_pHSTxtParser[1]->get_tokenpos();
      } else {
        tPos2 = g_pHSTxtParser[1]->get_tokenpos();
        while (token2 && tPos2 < tPos) {
          free (token2);
          token2  = g_pHSTxtParser[1]->next_token();
          tPos2   = g_pHSTxtParser[1]->get_tokenpos();
        }
      }
      bSkipToken2 = (!token2 || tPos2 > tPos) ? true : false;
      if (!bGood1 && token2 && !bSkipToken2)
        bGood2 = hunspell_spell(g_pHunSpell[1], token2) ? true : false;
    } // --> end of dict2

    if (!bGood1 && !bGood2) {
      if (!bGood2 && token2 && !bSkipToken2) {
        int pos = tPos2;
        if (g_codePage[1] == CP_UTF8)
          pos = ctCharsUTF8 (string_from_unicode_cp(start, g_codePage[1]), tPos2);
        ctLine = ctHardBreaks (bufW, 0, inspectSel.cpMin + pos);
        wSel.cpMin = inspectSel.cpMin + pos - ctLine;
        wSel.cpMax = wSel.cpMin + wcslen(string_to_unicode_cp(token2, g_codePage[1]));
      } else {
        int pos = tPos;
        if (g_codePage[0] == CP_UTF8)
          pos = ctCharsUTF8 (string_from_unicode_cp(start, g_codePage[0]), tPos);
        ctLine = ctHardBreaks (bufW, 0, inspectSel.cpMin + pos);
        wSel.cpMin = inspectSel.cpMin + pos - ctLine;
        wSel.cpMax = wSel.cpMin + wcslen(string_to_unicode_cp(token, g_codePage[0]));
      }
      SendMessage (hwnd, EM_EXSETSEL, 0, (LPARAM)&wSel);
      SendMessage (hwnd, EM_SETCHARFORMAT, SCF_SELECTION, (LPARAM)&g_badFormat);
      ctTypoz++;
    }
    free(token);
  } // while - for all tokens
  if (token)
    free(token);
  if (token2)
    free(token2);

  // go to prev selection
  SendMessage (hwnd, EM_EXSETSEL, 0, (LPARAM)&prevSel);
  // restore format at prev selection
  if (prevSel.cpMin == prevSel.cpMax)
    SendMessage (hwnd, EM_SETCHARFORMAT, SCF_SELECTION, (LPARAM)&g_goodFormat);
  // scroll to the same visible scroll area as we had before
  SendMessage (hwnd, EM_SETSCROLLPOS, 0, (LPARAM)&scrollPos);

  SendMessage (hwnd, WM_SETREDRAW, (WPARAM)TRUE, 0);
  InvalidateRect (hwnd, NULL, TRUE);
  UpdateWindow (hwnd);

  return ctTypoz;
}

_inline int suggest_word (HWND hwnd, MSGFILTER* msgf)
{

  if (!g_pHunSpell[0] || !g_pHSTxtParser[0] || !hwnd || !msgf)
    return 0;

  wchar_t     bufW[MAX_STRING_LEN];
  CHARFORMAT2 format;
  CHARRANGE   prevSel;
  RECT        rect;
  POINT       mouse;
  SendMessage (hwnd, EM_EXGETSEL, 0, (LPARAM)&prevSel);

  GetWindowRect (hwnd, &rect);
  mouse.x = LOWORD(msgf->lParam);
  mouse.y = HIWORD(msgf->lParam);

  int offset = LOWORD(SendMessage (hwnd, EM_CHARFROMPOS, 0, (LPARAM)&mouse));
  SendMessage (hwnd, EM_SETSEL, offset, offset);

  format.cbSize = sizeof (CHARFORMAT2);
  SendMessage (hwnd, EM_GETCHARFORMAT, 1, (LPARAM) &format);
  if ((format.dwEffects & CFU_UNDERLINEDOTTED) == 0) {
    SendMessage (hwnd, EM_EXSETSEL, 0, (LPARAM)&prevSel);
    return 0;
  }

  int sel_start, sel_end, selection;
  // find start of word
  sel_end = sel_start = offset;
  while (sel_start > 0) {
    sel_start--;
    SendMessage (hwnd, EM_SETSEL, (WPARAM) sel_start, (LPARAM) sel_start);
    SendMessage (hwnd, EM_GETCHARFORMAT, 1, (LPARAM) &format);
    if ((format.dwEffects & CFU_UNDERLINEDOTTED) == 0)
      break;
  }
  // find end of word
  while (sel_end < sel_start + (int)(_countof (bufW) - 1)) {
    sel_end++;
    SendMessage (hwnd, EM_SETSEL, (WPARAM) sel_end, (LPARAM) sel_end);
    SendMessage (hwnd, EM_GETSEL, (WPARAM) &selection, (WPARAM) &selection);
    SendMessage (hwnd, EM_GETCHARFORMAT, 1, (LPARAM) &format);
    if (selection < sel_end || (format.dwEffects & CFU_UNDERLINEDOTTED) == 0) {
      sel_end--;
      break;
    }
  }
  // get the text
  memset (bufW, 0, sizeof (bufW));
  TEXTRANGE textrg;
  textrg.lpstrText	= bufW;
  textrg.chrg.cpMin = sel_start;
  textrg.chrg.cpMax = sel_end;
  SendMessage (hwnd, EM_GETTEXTRANGE, 0, (LPARAM)&textrg);

  char*   token   = string_from_unicode_cp (bufW, g_codePage[0]);
  char*   token2  = string_from_unicode_cp (bufW, g_codePage[1]);
  char**  wlst    = NULL;
  int     ns      = hunspell_suggest(g_pHunSpell[0], token, &wlst);
  char**  wlst2   = NULL;
  int     ns2     = 0;
  if (g_pHunSpell[1] && token2)
    ns2 = hunspell_suggest(g_pHunSpell[1], token2, &wlst2);

  if ((ns + ns2) > 0) {
    UINT  cmdID = 0;
    HMENU hMenu = CreatePopupMenu ();

    if (ns)
      AppendMenu (hMenu, MF_GRAYED | MF_STRING, 9997, string_to_unicode (g_HSDictname));

    for (int j = 0; j < ns; j++, cmdID++)
      AppendMenu (hMenu, MF_STRING, cmdID + 1, string_to_unicode_cp (wlst[j], g_codePage[0]));

    if (ns2)
      AppendMenu (hMenu, MF_GRAYED | MF_STRING, 9998, string_to_unicode (g_HSDictname2));

    for (int j = 0; j < ns2; j++, cmdID++)
      AppendMenu (hMenu, MF_STRING, cmdID + 1, string_to_unicode_cp (wlst2[j], g_codePage[1]));

    POINT pt;
    GetCursorPos (&pt);
    UINT choice = TrackPopupMenu (hMenu, TPM_NONOTIFY | TPM_RETURNCMD, pt.x, pt.y, 0, hwnd, NULL);
    if (choice > 0 && choice <= (UINT)(ns + ns2)) {
      SendMessage (hwnd, EM_EXSETSEL, 0, (LPARAM)&textrg.chrg);
      SendMessage (hwnd, EM_SETCHARFORMAT, SCF_SELECTION, (LPARAM)&g_goodFormat);
      choice -= 1;
      if (choice >= 0 && choice < (UINT)ns)
        SendMessage (hwnd, EM_REPLACESEL, TRUE, (LPARAM)string_to_unicode_cp (wlst[choice], g_codePage[0]));
      else if ((choice - ns) >= 0 && (choice - ns) < (UINT)ns2)
        SendMessage (hwnd, EM_REPLACESEL, TRUE, (LPARAM)string_to_unicode_cp (wlst2[choice - ns], g_codePage[1]));
    }
    if (hMenu) {
      DestroyMenu(hMenu);
      hMenu = NULL;
    }
  }
  hunspell_free_list(g_pHunSpell[0], &wlst, ns);
  if (g_pHunSpell[1])
    hunspell_free_list(g_pHunSpell[1], &wlst2, ns2);

  SendMessage (hwnd, EM_EXSETSEL, 0, (LPARAM)&prevSel);

  return (ns + ns2);
}

int spellchecker_process (HWND hwnd, MSGFILTER* msgf)
{
  if (!hwnd || !msgf)
    return 0;

  if (msgf->msg == WM_RBUTTONUP)
    return suggest_word (hwnd, msgf);
  else if (msgf->msg == WM_LBUTTONDOWN
  || msgf->msg == WM_MBUTTONDOWN
  || msgf->msg == WM_RBUTTONDOWN
  || msgf->msg == WM_XBUTTONDOWN)
    return check_spelling (hwnd, msgf, -1);
  else if (msgf->msg == WM_KEYUP
  && (
    msgf->wParam == VK_LEFT ||
    msgf->wParam == VK_RIGHT ||
    msgf->wParam == VK_UP ||
    msgf->wParam == VK_DOWN ||
    msgf->wParam == VK_DELETE ||
    msgf->wParam == VK_INSERT ||
    msgf->wParam == VK_HOME ||
    msgf->wParam == VK_END ||
    msgf->wParam == VK_PRIOR ||
    msgf->wParam == VK_NEXT ||
    msgf->wParam == VK_TAB ||
    msgf->wParam == VK_BACK ||
    msgf->wParam == VK_RETURN
  )) {
    return check_spelling (hwnd, msgf, -1);
  } else if (msgf->msg == WM_CHAR && isEndWordWchar(msgf->wParam))
    return check_spelling (hwnd, msgf, 16);

  return 0;
}

_inline int cleanup_checker (DWORD num)
{
  if (num >= DICT_MAX)
    return 0;
  if (g_pHunSpell[num]) {
    hunspell_uninitialize(g_pHunSpell[num]);
    g_pHunSpell[num] = NULL;
    debug_log ("Hunspell dictionary #%d %s unloaded succesfully\n",
      num + 1, num ==  0 ? g_HSDictname : g_HSDictname2);
  }
  if (g_pHSTxtParser[num]) {
    delete g_pHSTxtParser[num];
    g_pHSTxtParser[num] = NULL;
  }
  if (g_dicEnc[num]) {
    free (g_dicEnc[num]);
    g_dicEnc[num] = NULL;
  }
  return 0;
}

char* spellchecker_dictname_get (void)
{
  return g_HSDictname;
}

char* spellchecker_dictname2_get (void)
{
  return g_HSDictname2;
}

int spellchecker_dictname_set (char* dictname)
{
  if (!dictname || !dictname[0]
  || (!strcmp (dictname, DICT_DISABLED))) {
    strncpy_s (g_HSDictname, _countof(g_HSDictname), DICT_DISABLED, _TRUNCATE);
    return spellchecker_init ();
  }
  if (strcmp (g_HSDictname, dictname)) {
    strncpy_s (g_HSDictname, _countof(g_HSDictname), dictname, _TRUNCATE);
    return spellchecker_init ();
  }
  return 0;
}

int spellchecker_dictname2_set (char* dictname2)
{
  if (!dictname2 || !dictname2[0]
  || (!strcmp (dictname2, DICT_DISABLED))) {
    strncpy_s (g_HSDictname2, _countof(g_HSDictname2), DICT_DISABLED, _TRUNCATE);
    cleanup_checker (1);
    return 0;
  }
  if (strcmp (g_HSDictname2, dictname2)) {
    strncpy_s (g_HSDictname2, _countof(g_HSDictname2), dictname2, _TRUNCATE);
    return spellchecker_init ();
  }
  return 0;
}

_inline bool dict_loader (int d, char* dic, char* aff)
{
  char* dic_enc = NULL;  // dictionary encoding == codepage name

  g_pHunSpell[d] = (Hunspell*)hunspell_initialize (aff, dic);
  if (!g_pHunSpell[d])
    return false;
  dic_enc = hunspell_get_dic_encoding(g_pHunSpell[d]);
  if (!dic_enc || !dic_enc[0] || strlen(g_HSDictname) < 2) {
    delete g_pHunSpell[d];
    g_pHunSpell[d] = NULL;
    debug_log ("Hunspell dictionary #%d %s failed loading\n",
      d + 1, d == 0 ? g_HSDictname : g_HSDictname2);
    return false;
  }
  g_dicEnc[d]   = _strdup (dic_enc);
  g_codePage[d] = string_getCodePage (dic_enc);
  char eTmp[3]  = { (char)tolower(g_HSDictname[0] & 0xff), (char)tolower(g_HSDictname[1] & 0xff), (char)0x0, };
  g_langNum[d]  = get_lang_num (eTmp);
  if (g_langNum[d] == LANG_xx)
    g_langNum[d] = g_pHunSpell[d]->get_langnum();

  char*           wordchars           = 0;
  int             wordchars_utf16_len = 0;
  unsigned short* wordchars_utf16     = 0;

  if (!strcmp (dic_enc, "UTF-8") || !strcmp (dic_enc, "UTF8")) {
    wordchars_utf16 = g_pHunSpell[d]->get_wordchars_utf16(&wordchars_utf16_len);
    initialize_utf_tbl();
  } else {
    char* casechars = get_casechars(dic_enc);
    wordchars = (char*)g_pHunSpell[d]->get_wordchars();
    if (casechars && wordchars) {
      int sz = strlen (casechars) + strlen(wordchars) + 1;
      casechars = (char*)realloc(casechars, sz);
      strcat_s (casechars, sz, wordchars);
    }
    wordchars = casechars;
  }
  if (wordchars)
    g_pHSTxtParser[d] = new TextParser(wordchars);
  else
    g_pHSTxtParser[d] = new TextParser(wordchars_utf16, wordchars_utf16_len);

  g_pHSTxtParser[d]->set_url_checking(TRUE);

  debug_log ("Hunspell dictionary #%d %s code page %d %s loaded succesfully\n", d + 1,
    d == 0 ? g_HSDictname : g_HSDictname2,
    g_codePage[d], g_dicEnc[d]);

  return true;
}

int spellchecker_init (void)
{
  spellchecker_term ();

  char  dic[512];
  char  aff[512];

  g_goodFormat.cbSize         = sizeof (g_goodFormat);
  g_goodFormat.dwMask         = CFM_UNDERLINE | CFM_UNDERLINETYPE | CFM_BACKCOLOR | CFM_COLOR;
  g_goodFormat.dwEffects      = CFE_AUTOBACKCOLOR | CFE_AUTOCOLOR;
  g_goodFormat.bUnderlineType = CFU_UNDERLINENONE;
  g_goodFormat.crBackColor    = RGB(255, 255, 255);
  g_goodFormat.crTextColor    = RGB(0, 0, 0);

  g_urlFormat.cbSize          = sizeof (g_urlFormat);
  g_urlFormat.dwMask          = CFM_UNDERLINE | CFM_UNDERLINETYPE | CFM_BACKCOLOR | CFM_COLOR;
  g_urlFormat.dwEffects       = CFE_UNDERLINE | CFE_AUTOBACKCOLOR;
  g_urlFormat.bUnderlineType  = CFU_UNDERLINE;
  g_urlFormat.crBackColor     = RGB(255, 255, 255);
  g_urlFormat.crTextColor     = RGB(127, 127, 159); // greyed out

  g_badFormat.cbSize         = sizeof (g_badFormat);
  g_badFormat.dwMask         = CFM_UNDERLINE | CFM_UNDERLINETYPE | CFM_BACKCOLOR | CFM_COLOR;
  g_badFormat.dwEffects      = CFE_UNDERLINE | CFE_AUTOBACKCOLOR;
  g_badFormat.bUnderlineType = CFU_UNDERLINEDOTTED;
  g_badFormat.crBackColor    = RGB(255, 255, 255);
  g_badFormat.crTextColor    = RGB(192, 64, 64);


  if (strcmp (g_HSDictname, DICT_DISABLED)) {
    if (!g_HSDictname[0]) {
      strcpy_s (g_HSDictname, _countof(g_HSDictname), "en_US");
      debug_log ("Hunspell dictionary #%d %s using default.\n", 1, g_HSDictname);
    }
    sprintf_s (dic, _countof (dic), "%s/%s.dic", DICT_PATH, g_HSDictname);
    sprintf_s (aff, _countof (aff), "%s/%s.aff", DICT_PATH, g_HSDictname);
    if (GetFileAttributesA (dic) == -1 || GetFileAttributesA (aff) == -1) {
      if (!strcmp (g_HSDictname, "en_US")) {
        debug_log ("Hunspell dictionary #%d %s failed, file(s) missing\n", 1, g_HSDictname);
      } else {
        debug_log ("Hunspell dictionary #%d %S failed falling back to en_US\n", 1, g_HSDictname);
        strcpy_s (g_HSDictname, _countof (g_HSDictname), "en_US");
        sprintf_s (dic, _countof (dic), "%s/%s.dic", DICT_PATH, g_HSDictname);
        sprintf_s (aff, _countof (aff), "%s/%s.aff", DICT_PATH, g_HSDictname);
        if (GetFileAttributesA (dic) == -1 || GetFileAttributesA (aff) == -1)
          debug_log ("Hunspell dictionary #%d %s failed, file(s) missing\n", 1, g_HSDictname);
        else
          g_availabe[0] = true;
      }
    } else {
      g_availabe[0] = true;
    }
  }

  if (!g_HSDictname2[0])
    strcpy_s (g_HSDictname2, _countof(g_HSDictname2), DICT_DISABLED);
  else if (strcmp (g_HSDictname2, DICT_DISABLED)) {
    sprintf_s (dic, _countof (dic), "%s/%s.dic", DICT_PATH, g_HSDictname2);
    sprintf_s (aff, _countof (aff), "%s/%s.aff", DICT_PATH, g_HSDictname2);
    if (GetFileAttributesA (dic) != -1 && GetFileAttributesA (aff) != -1)
      g_availabe[1] = true;
  }

  if (g_availabe[0]) {
    debug_log ("Hunspell dictionary #%d %s available\n", 1, g_HSDictname);
    if (g_availabe[1])
      debug_log ("Hunspell dictionary #%d %s available\n", 2, g_HSDictname2);
  } else if (g_availabe[1]) {
    char dictname[32];
    // dict1 not avalaible, loading dict2 only
    // swap names:
    strcpy_s (dictname, _countof(dictname), g_HSDictname);
    strcpy_s (g_HSDictname, _countof(g_HSDictname), g_HSDictname2);
    strcpy_s (g_HSDictname2, _countof(g_HSDictname2), dictname);
    // swap available:
    g_availabe[0] = g_availabe[1];
    g_availabe[1] = false;
    debug_log ("Hunspell dictionary #%d %s available\n", 1, g_HSDictname);
  } else {
    debug_log ("Spell checking disabled, or no dictionaries found.\n");
    return 0;
  }

  sprintf_s (dic, _countof (dic), "%s/%s.dic", DICT_PATH, g_HSDictname);
  sprintf_s (aff, _countof (aff), "%s/%s.aff", DICT_PATH, g_HSDictname);
  dict_loader (0, dic, aff);

  if (g_availabe[1]) {
    if (!strcmp (g_HSDictname, g_HSDictname2)) {
      // don't need a second copy running...
      strncpy_s (g_HSDictname2, _countof(g_HSDictname2), DICT_DISABLED, _TRUNCATE);
    } else {
      sprintf_s (dic, _countof (dic), "%s/%s.dic", DICT_PATH, g_HSDictname2);
      sprintf_s (aff, _countof (aff), "%s/%s.aff", DICT_PATH, g_HSDictname2);
      dict_loader (1, dic, aff);
    }
  }

  return (g_availabe[0] + g_availabe[1]);
}

int spellchecker_term (void)
{
  for (int d = 0; d < DICT_MAX; d++) {
    cleanup_checker (d);
    g_codePage[d] = CP_ACP;
    g_availabe[d] = false;
  }
  free_utf_tbl();
  return 0;
}